from dataclasses import dataclass
from datetime import date
from typing import Dict, List, Optional, Tuple, Union

from marshmallow import Schema, fields, post_dump

# Path parts known to indicate the level of law they refer to.
KNOWN_CODES = {
    "test": {"acts": ["Test Acts", "S"]},
    "us": {
        "const": ["U.S. Const.", "C"],
        "usc": ["U.S. Code", "S"],
        "cfr": ["CFR", "R"],
    },
}

CODE_LEVEL_NAMES = {"C": "constitution", "S": "statute", "R": "regulation"}


def identify_code(jurisdiction: str, code: str) -> Tuple[str, str]:
    """Find code name and type based on USLM citation parts."""
    try:
        sovereign = KNOWN_CODES[jurisdiction]
    except KeyError:
        raise KeyError(f'"{jurisdiction}" is not a known jurisdiction identifier')

    try:
        code_name, code_level = sovereign[code]
    except KeyError:
        raise KeyError(f'"{code}" is not a known code identifier')

    code_level_name = CODE_LEVEL_NAMES[code_level]
    return code_name, code_level_name


class CitationSchema(Schema):
    document_type = fields.Str(data_key="type", default="legislation", dump_only=True)
    jurisdiction = fields.Str(required=True)
    code = fields.Str(data_key="container-title", required=False)
    volume = fields.Str(required=False)
    section = fields.Str(required=False)
    revision_date = fields.Date(default=None, load_only=True)
    event_date = fields.Method(
        data_key="event-date", serialize="dump_event_date", dump_only=True
    )

    class Meta:
        ordered = True

    def dump_event_date(self, obj) -> Optional[Dict[str, List[List[Union[str, int]]]]]:
        if not obj.revision_date:
            return None

        return {
            "date-parts": [
                [
                    str(obj.revision_date.year),
                    obj.revision_date.month,
                    obj.revision_date.day,
                ]
            ]
        }


@dataclass
class Citation:
    r"""
    A citation style for referring to an :class:`~legislice.enactments.Enactment` in written text.

    Intended for use with `Citation Style Language (CSL) <https://citeproc-js.readthedocs.io/en/latest/csl-json/markup.html>`_.
    """

    def __init__(
        self,
        jurisdiction: str,
        code: Optional[str] = None,
        volume: Optional[str] = None,
        section: Optional[str] = None,
        revision_date: Optional[date] = None,
    ) -> None:
        self.jurisdiction = jurisdiction

        if code:
            code, code_level_name = identify_code(jurisdiction=jurisdiction, code=code)
        self.code = code

        if volume:
            volume = volume.lstrip("t")
        self.volume = volume

        if section:
            section = "sec. " + section.lstrip("s")
        self.section = section

        self.revision_date = revision_date

    def __str__(self):
        name = f"{self.volume} {self.code} {self.section}"
        if self.revision_date:
            name += f" ({self.revision_date.year})"
        return name.replace("sec.", "§")

    def as_dict(self) -> str:
        """Dump Citation Style Language data as Python object."""
        schema = CitationSchema()
        return schema.dump(self)

    def as_json(self) -> str:
        """Dump Citation Style Language data as JSON."""
        schema = CitationSchema()
        return schema.dumps(self)



"""EnactmentGroup class."""

from __future__ import annotations

import textwrap
from typing import List, Optional, Sequence, Tuple, Union

from legislice.enactments import Enactment, consolidate_enactments


class EnactmentGroup:
    """Group of Enactments with comparison methods."""

    def __init__(
        self,
        enactments: Optional[
            Union[EnactmentGroup, Sequence[Enactment], Enactment]
        ] = None,
    ):
        """Normalize ``factors`` as sequence attribute."""
        if isinstance(enactments, EnactmentGroup):
            self.sequence: List[Enactment] = enactments.sequence
        elif isinstance(enactments, Sequence):
            self.sequence = list(enactments)
        elif enactments is None:
            self.sequence = []
        else:
            self.sequence = [enactments]
        for enactment in self.sequence:
            if not isinstance(enactment, Enactment):
                raise TypeError(
                    f'Object "{enactment} could not be included in '
                    f"{self.__class__.__name__} because it is "
                    f"type {enactment.__class__.__name__}, not type Enactment"
                )
        self.sequence = consolidate_enactments(self.sequence)
        self.sequence.sort(key=lambda x: x.node)

    def _at_index(self, key: int) -> Enactment:
        return self.sequence[key]

    def __getitem__(self, key: Union[int, slice]) -> Enactment:
        if isinstance(key, slice):
            start, stop, step = key.indices(len(self))
            return self.__class__([self._at_index(i) for i in range(start, stop, step)])
        return self._at_index(key)

    def __iter__(self):
        yield from self.sequence

    def __len__(self):
        return len(self.sequence)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({repr(list(self.sequence))})"

    def __str__(self):
        result = "the group of Enactments:"
        indent = "  "
        for factor in self.sequence:
            result += f"\n{textwrap.indent(str(factor), prefix=indent)}"
        return result

    def _add_group(self, other: EnactmentGroup) -> EnactmentGroup:
        combined = self.sequence[:] + other.sequence[:]
        return self.__class__(combined)

    def __add__(
        self, other: Union[EnactmentGroup, Sequence[Enactment], Enactment]
    ) -> EnactmentGroup:
        """Combine two EnactmentGroups, consolidating any duplicate Enactments."""
        if isinstance(other, self.__class__):
            return self._add_group(other)
        to_add = self.__class__(other)
        return self._add_group(to_add)

    def __ge__(self, other: Union[Enactment, EnactmentGroup]) -> bool:
        """Test whether ``self`` implies ``other`` and ``self`` != ``other``."""
        return bool(self.implies(other))

    def __gt__(self, other: Union[Enactment, EnactmentGroup]) -> bool:
        """Test whether ``self`` implies ``other`` and ``self`` != ``other``."""
        return bool(self.implies(other))

    def _implies_enactment(self, other: Enactment) -> bool:
        return any(self_enactment.implies(other) for self_enactment in self)

    def _implies(self, other: EnactmentGroup) -> bool:
        return all(self._implies_enactment(other_law) for other_law in other)

    def implies(self, other: Union[Enactment, EnactmentGroup]) -> bool:
        """Determine whether self includes all the text of another Enactment or EnactmentGroup."""
        if isinstance(other, Enactment):
            return self._implies_enactment(other)
        return self._implies(other)



import pytest


class TestUpdateEnactments:
    """
    Tests for filling in blank fields in JSON Enactments with the test Client.

    The Enactments that need filling in could be user-generated as part of a model
    of an AuthoritySpoke holding, and can have some fields missing.
    The Enactments that come from the Client can be assumed not to be missing fields.
    """

    @pytest.mark.vcr
    def test_update_entries_in_enactment_index(self, test_client):
        enactment_index = {
            "security": {
                "node": "/us/const/amendment/IV",
                "start_date": "1791-12-15",
                "exact": "right of the people to be secure",
            }
        }
        client = test_client
        updated_index = client.update_entries_in_enactment_index(enactment_index)
        updated_enactment = updated_index["security"]
        assert updated_enactment["heading"] == "AMENDMENT IV."
        assert updated_enactment["url"].startswith("https")

    @pytest.mark.vcr
    def test_update_entry_without_date(self, test_client):
        enactment_index = {
            "person clause": {
                "node": "/us/const/amendment/XIV/3",
                "start": 3,
                "end": 15,
            }
        }
        client = test_client
        updated_index = client.update_entries_in_enactment_index(enactment_index)
        updated_enactment = updated_index["person clause"]
        assert updated_enactment["heading"].startswith("Loyalty as a qualification")
        loaded_enactment = client.read_from_json(updated_enactment)
        assert loaded_enactment.selected_text() == "…person shall…"

    @pytest.mark.vcr
    def test_read_enactment_with_suffix_field(self, test_client):
        raw_enactment = {
            "name": "search clause",
            "node": "/us/const/amendment/IV",
            "suffix": ", and no Warrants shall issue",
            "start_date": "1791-12-15",
        }
        enactment = test_client.read_from_json(raw_enactment)
        assert enactment.selected_text().endswith("shall not be violated…")

    @pytest.mark.vcr()
    def test_update_linked_enactment(self, test_client):
        data = {"node": "/us/const"}
        new = test_client.update_enactment_from_api(data)
        assert new["node"] == "/us/const"
        assert new["start_date"] == "1788-09-13"
        assert isinstance(new["children"][0], str)

    @pytest.mark.vcr()
    def test_update_enactment_when_reading_from_json(self, test_client):
        enactment = test_client.read_from_json(data={"node": "/us/const/amendment/IV"})
        assert enactment.start_date.isoformat() == "1791-12-15"

    @pytest.mark.vcr()
    def test_text_in_updated_enactment_is_selected_by_default(self, test_client):
        client = test_client
        enactment = client.read_from_json(data={"node": "/us/const/amendment/IV"})
        assert enactment.selected_text().startswith("The right")



from copy import deepcopy

import pytest

from legislice.groups import EnactmentGroup


class TestEnactmentGroups:
    def test_make_group(self, copyright_clause, copyright_statute):
        group = EnactmentGroup([copyright_clause, copyright_statute])
        assert len(group) == 2
        assert isinstance(EnactmentGroup(group), EnactmentGroup)

    def test_consolidate_adjacent_passages(self, copyright_clause, copyright_statute):
        copyright_clause.select(None)
        securing_for_authors = copyright_clause + (
            "To promote the Progress of Science and "
            "useful Arts, by securing for limited Times to Authors"
        )
        and_inventors = copyright_clause + "and Inventors"
        right_to_writings = (
            copyright_clause + "the exclusive Right to their respective Writings"
        )
        left = EnactmentGroup([and_inventors, copyright_statute])
        right = EnactmentGroup([securing_for_authors, right_to_writings])

        combined = left + right
        assert len(combined) == 2
        assert "/us/usc/t17/s102/b" in repr(combined)
        assert any(
            law.selected_text().startswith("To promote the Progress")
            and law.selected_text().endswith("their respective Writings…")
            for law in combined
        )

    def test_wrong_type_in_group(self, section6d, test_client):
        section = test_client.read_from_json(section6d)
        cite = section.as_citation()
        with pytest.raises(TypeError):
            EnactmentGroup([cite])


class TestImplies:
    def test_no_implication_of_group(self, copyright_clause, copyright_statute):
        extra = deepcopy(copyright_clause)
        extra.select(None)
        securing_for_authors = extra + (
            "To promote the Progress of Science and "
            "useful Arts, by securing for limited Times to Authors"
        )
        and_inventors = extra + "and Inventors"
        left = EnactmentGroup(copyright_clause)
        right = EnactmentGroup([securing_for_authors, and_inventors, copyright_statute])
        assert not left.implies(right)

    def test_implication_of_group(self, copyright_clause, copyright_statute):
        extra = deepcopy(copyright_clause)
        extra.select(None)
        securing_for_authors = extra + (
            "To promote the Progress of Science and "
            "useful Arts, by securing for limited Times to Authors"
        )
        and_inventors = extra + "and Inventors"
        left = EnactmentGroup([copyright_clause, copyright_statute])
        right = EnactmentGroup([securing_for_authors, and_inventors])
        assert left.implies(right)
        assert left >= right

    def test_implication_of_enactment(self, copyright_clause, copyright_statute):
        extra = deepcopy(copyright_clause)
        extra.select(None)
        securing_for_authors = extra + (
            "To promote the Progress of Science and "
            "useful Arts, by securing for limited Times to Authors"
        )
        left = EnactmentGroup([copyright_clause, copyright_statute])
        assert left > securing_for_authors


class TestAdd:
    def test_add_enactment_to_group(self, copyright_clause):
        copyright_clause.select(None)
        securing_for_authors = copyright_clause + (
            "To promote the Progress of Science and "
            "useful Arts, by securing for limited Times to Authors"
        )
        and_inventors = copyright_clause + "and Inventors"
        right_to_writings = (
            copyright_clause + "the exclusive Right to their respective Writings"
        )
        left = EnactmentGroup([securing_for_authors, and_inventors])
        right = right_to_writings
        result = left + right
        assert len(result) == 1
        assert "respective Writings…" in str(result)
        assert "/us/const/article/I/8/8" in str(result[0])
        assert "1788-09-13" in str(result[:])

    def test_add_empty_group(self, copyright_clause, copyright_statute):
        copyright_clause.select(None)
        securing_for_authors = copyright_clause + (
            "To promote the Progress of Science and "
            "useful Arts, by securing for limited Times to Authors"
        )
        writings = copyright_clause + "their respective Writings"
        left = EnactmentGroup([securing_for_authors, writings, copyright_statute])
        assert len(left) == 2
        right = EnactmentGroup(None)
        result = left + right
        assert len(result) == 2

    def test_enactments_ordered_after_adding_groups(self, first_a, second_a, third_a):
        establishment_clause = deepcopy(first_a)
        establishment_clause.select(
            "Congress shall make no law respecting an establishment of religion"
        )
        speech_clause = deepcopy(first_a)
        speech_clause.select(
            ["Congress shall make no law", "abridging the freedom of speech"]
        )

        arms_clause = deepcopy(second_a)
        arms_clause.select(
            "the right of the people to keep and bear arms, shall not be infringed."
        )
        third_amendment = deepcopy(third_a)

        left = EnactmentGroup([establishment_clause, arms_clause])
        right = EnactmentGroup([third_amendment, speech_clause])

        combined = left + right
        assert len(combined) == 3
        assert combined[0].node == "/us/const/amendment/I"



from copy import deepcopy
from datetime import datetime
import json

import os

from anchorpoint.textselectors import TextPositionSelector, TextQuoteSelector
from dotenv import load_dotenv
from marshmallow import ValidationError
import pytest

from legislice.download import Client

from legislice.schemas import (
    CrossReferenceSchema,
    EnactmentSchema,
    InboundReferenceSchema,
    LinkedEnactmentSchema,
    SelectorSchema,
    CitingProvisionLocationSchema,
    enactment_needs_api_update,
)

load_dotenv()

TOKEN = os.getenv("LEGISLICE_API_TOKEN")
API_ROOT = os.getenv("API_ROOT")


class TestLoadSelector:
    def test_schema_loads_position_selector(self):
        schema = SelectorSchema()
        data = {"start": 0, "end": 12}
        result = schema.load(data)
        assert isinstance(result, TextPositionSelector)

    def test_selector_text_split(self):
        schema = SelectorSchema()
        data = {"text": "process, system,|method of operation|, concept, principle"}
        result = schema.load(data)
        assert result.exact.startswith("method")

    def test_selector_from_string(self):
        data = "eats,|shoots,|and leaves"
        schema = SelectorSchema()
        result = schema.load(data)
        assert result.exact == "shoots,"

    def test_selector_from_string_without_split(self):
        data = "promise me not to omit a single word"
        schema = SelectorSchema()
        result = schema.load(data)
        assert result.exact.startswith("promise")

    def test_selector_from_string_split_wrongly(self):
        data = "eats,|shoots,|and leaves|"
        schema = SelectorSchema()
        with pytest.raises(ValidationError):
            _ = schema.load(data)


class TestLoadCrossReference:
    client = Client(api_token=TOKEN, api_root=API_ROOT)

    def test_load_citation(self, citation_to_6c):
        schema = CrossReferenceSchema()
        result = schema.load(citation_to_6c)
        assert result.target_uri == "/test/acts/47/6C"
        assert (
            str(result)
            == 'CrossReference(target_uri="/test/acts/47/6C", reference_text="Section 6C")'
        )


class TestLoadEnactment:
    client = Client(api_token=TOKEN, api_root=API_ROOT)

    def test_load_nested_enactment(self, section6d):
        schema = EnactmentSchema()
        result = schema.load(section6d)
        assert result.heading.startswith("Waiver")

    def test_enactment_with_nested_selectors(self, section_11_subdivided):
        schema = EnactmentSchema()
        section_11_subdivided["selection"] = [{"start": 0}]
        for child in section_11_subdivided["children"]:
            child["selection"] = []
        section_11_subdivided["children"][1]["selection"] = [{"start": 0, "end": 12}]
        result = schema.load(section_11_subdivided)
        answer = "The Department of Beards may issue licenses to such…hairdressers…"
        assert result.selected_text() == answer

    def test_enactment_with_True_as_selector(self, section_11_subdivided):
        schema = EnactmentSchema()
        section_11_subdivided["selection"] = True
        section_11_subdivided["children"][1]["selection"] = [{"start": 0, "end": 12}]
        result = schema.load(section_11_subdivided)
        answer = "The Department of Beards may issue licenses to such…hairdressers…"
        assert result.selected_text() == answer

    def test_enactment_with_False_as_selector(self, section_11_subdivided):
        schema = EnactmentSchema()
        section_11_subdivided["selection"] = False
        section_11_subdivided["children"][1]["selection"] = [{"start": 0, "end": 12}]
        result = schema.load(section_11_subdivided)
        answer = "…hairdressers…"
        assert result.selected_text() == answer

    def test_selector_not_wrapped_in_list(self, section_11_together):
        schema = EnactmentSchema()
        section_11_together["selection"] = {"start": 4, "end": 24}
        result = schema.load(section_11_together)
        assert result.selected_text() == "…Department of Beards…"

    def test_load_with_text_quote_selector(self, section_11_together):
        schema = EnactmentSchema()
        section_11_together["selection"] = [{"exact": "Department of Beards"}]
        result = schema.load(section_11_together)
        assert result.selected_text() == "…Department of Beards…"

    @pytest.mark.vcr()
    def test_load_enactment_with_text_anchor(
        self, provision_with_text_anchor, test_client
    ):
        schema = EnactmentSchema()
        record = test_client.update_enactment_from_api(provision_with_text_anchor)
        result = schema.load(record)
        assert result.anchors[0].exact == "17 U.S.C. § 102(a)"

    def test_enactment_does_not_fail_for_excess_selector(self, section_11_subdivided):
        """Test selector that extends into the text of a subnode."""
        exact = "The Department of Beards may issue licenses to such barbers"
        section_11_subdivided["exact"] = exact
        schema = EnactmentSchema(many=False)
        enactment = schema.load(section_11_subdivided)

        assert enactment.selected_text() == exact + "…"

    def test_node_field_needed_to_load_enactment(self):
        barbers_without_node = {
            "heading": "",
            "content": "barbers,",
            "children": [],
            "end_date": None,
            "start_date": "2013-07-18",
            "url": "https://authorityspoke.com/api/v1/test/acts/47/11/i@2020-01-01",
        }
        with pytest.raises(ValueError):
            _ = enactment_needs_api_update(barbers_without_node)

    def test_nest_selector_fields_before_loading(self, test_client, fourth_a):
        fourth_a["selection"] = [{"start": 10, "end": 20}]
        fourth_a["suffix"] = ", and no Warrants shall issue"
        schema = EnactmentSchema()
        updated = schema.move_selector_fields(fourth_a)
        assert updated["selection"][1]["suffix"] == ", and no Warrants shall issue"

    def test_load_enactment_with_cross_reference(self):
        data = {
            "heading": "",
            "content": "The Department of Beards shall waive the collection of beard tax upon issuance of beardcoin under Section 6C where the reason the maintainer wears a beard is due to bona fide religious, cultural, or medical reasons.",
            "start_date": "2013-07-18",
            "node": "/test/acts/47/6D/1",
            "children": [],
            "end_date": None,
            "url": "http://127.0.0.1:8000/api/v1/test/acts/47/6D/1@2020-01-01/",
            "citations": [
                {
                    "target_uri": "/test/acts/47/6C",
                    "target_url": "http://127.0.0.1:8000/api/v1/test/acts/47/6C@2020-01-01/",
                    "target_node": 1660695,
                    "reference_text": "Section 6C",
                }
            ],
        }
        schema = EnactmentSchema(many=False)
        enactment = schema.load(data)
        assert len(enactment._cross_references) == 1

    def test_load_enactment_with_nested_cross_reference(self):
        data = {
            "heading": "Waiver of beard tax in special circumstances",
            "content": "",
            "start_date": "1935-04-01",
            "node": "/test/acts/47/6D",
            "children": [
                {
                    "heading": "",
                    "content": "The Department of Beards shall waive the collection of beard tax upon issuance of beardcoin under Section 6C where the reason the maintainer wears a beard is due to bona fide religious, cultural, or medical reasons.",
                    "start_date": "2013-07-18",
                    "node": "/test/acts/47/6D/1",
                    "children": [],
                    "end_date": None,
                    "url": "http://127.0.0.1:8000/api/v1/test/acts/47/6D/1@2020-01-01",
                    "citations": [
                        {
                            "target_uri": "/test/acts/47/6C",
                            "target_url": "http://127.0.0.1:8000/api/v1/test/acts/47/6C@2020-01-01",
                            "target_node": 1660695,
                            "reference_text": "Section 6C",
                        }
                    ],
                }
            ],
            "end_date": None,
            "url": "http://127.0.0.1:8000/api/v1/test/acts/47/6D@2020-01-01",
            "citations": [],
            "parent": "http://127.0.0.1:8000/api/v1/test/acts/47@2020-01-01",
        }
        schema = EnactmentSchema(many=False)
        enactment = schema.load(data)
        refs = enactment.children[0]._cross_references
        assert len(refs) == 1


class TestLoadLinkedEnactment:
    client = Client(api_token=TOKEN, api_root=API_ROOT)

    def test_load_linked_enactment(self):
        schema = LinkedEnactmentSchema()
        data = {
            "type": "Enactment",
            "children": [
                "https://authorityspoke.com/api/v1/us/const/",
                "https://authorityspoke.com/api/v1/us/usc/",
            ],
            "content": "",
            "end_date": None,
            "heading": "United States Legislation",
            "node": "/us",
            "parent": None,
            "start_date": "1776-07-04",
            "url": "https://authorityspoke.com/api/v1/us/const/",
        }
        result = schema.load(data)
        assert result.children[0] == "https://authorityspoke.com/api/v1/us/const/"

    @pytest.mark.vcr
    def test_text_sequence_for_linked_enactment(self, test_client):
        schema = LinkedEnactmentSchema()
        enactment = test_client.read(query="/test", date="2020-01-01")
        assert "for documentation." in enactment.text_sequence()[0].text
        enactment.select("for documentation.")
        assert enactment.selected_text() == "…for documentation."
        dumped = schema.dump(enactment)
        loaded = schema.load(dumped)
        assert loaded.selected_text() == "…for documentation."


class TestDumpEnactment:
    @pytest.mark.vcr()
    def test_dump_enactment_with_selector_to_dict(self, test_client):
        copyright_clause = test_client.read("/us/const/article/I/8/8")
        copyright_clause.select("Science and useful Arts")

        schema = EnactmentSchema()
        dumped = schema.dump(copyright_clause)
        selection = dumped["selection"][0]
        quote = dumped["text_version"]["content"][selection["start"] : selection["end"]]
        assert quote == "Science and useful Arts"

    @pytest.mark.vcr()
    def test_serialize_enactment_after_adding(self, test_client):
        search = test_client.read("/us/const/amendment/IV")
        warrant = deepcopy(search)

        search.select(TextQuoteSelector(suffix=", and no Warrants"))
        warrant.select(
            TextQuoteSelector(
                exact="shall not be violated, and no Warrants shall issue,"
            )
        )
        combined_enactment = search + warrant

        # search.select_more_text_at_current_node(warrant.selection)

        schema = EnactmentSchema()
        dumped = schema.dump(combined_enactment)

        assert dumped["text_version"]["content"].startswith("The right")
        assert dumped["text_version"].get("uri") is None

    @pytest.mark.vcr()
    def test_fields_ordered_with_children_last(self, test_client):
        s103 = test_client.read(query="/us/usc/t17/s103", date="2020-01-01")
        schema = EnactmentSchema()
        dumped = schema.dump(s103)
        assert list(dumped["children"][0]["selection"][0].keys()) == ["start", "end"]
        as_json = json.dumps(dumped)
        # Start field comes before end field in selector
        assert '"selection": [{"start":' in as_json
        # "Children" field is last, since it's hard to read otherwise
        assert list(dumped.keys())[-1] == "children"


class TestLoadInboundReferences:
    def test_load_citing_location_without_heading(self):
        location = {
            "heading": "",
            "node": "/us/usc/t17/s109/b/4",
            "start_date": "2013-07-18",
        }
        schema = CitingProvisionLocationSchema()
        loaded = schema.load(location)
        assert loaded.heading == ""
        assert loaded.start_date.isoformat() == "2013-07-18"

    def test_load_inbound_reference(self):
        """
        Test assuming the real target of the citation is given at the outer level of the dict.

        Only the first citation should be used to get data for the loaded object. The other
        one is irrelevant.
        """
        data = {
            "target_uri": "/us/usc/t17/s501",
            "citations": [
                {
                    "reference_text": "section 501 of this title",
                    "target_node": 1252985,
                    "target_uri": "/us/usc/t17/s501",
                    "target_url": "http://127.0.0.1:8000/api/v1/us/usc/t17/s501/",
                },
                {
                    "reference_text": "section 2319 of title 18",
                    "target_node": 1151186,
                    "target_uri": "/us/usc/t18/s2319",
                    "target_url": "http://127.0.0.1:8000/api/v1/us/usc/t18/s2319/",
                },
            ],
            "content": "Any person who distr... title 18.",
            "locations": [
                {
                    "heading": "",
                    "node": "/us/usc/t17/s109/b/4",
                    "start_date": "2013-07-18",
                }
            ],
            "url": "http://127.0.0.1:8000/api/v1/textversions/1031604/",
        }
        schema = InboundReferenceSchema()
        loaded = schema.load(data)
        assert loaded.target_uri == "/us/usc/t17/s501"
        assert loaded.reference_text == "section 501 of this title"
